// REQUIRES: common.js, webvisum.js, protocol.js



/*
 * Manages the context for execution of user operations (labeling, ocr, captcha).
 * Is a singleton per local webvisum. Generates/prepares Webvisum.ContextManager.Context for each operation.  
 */
Webvisum.ContextManager = {
    
    init : function(aLocalWebvisum) {
        this.webvisum = aLocalWebvisum;
        this.listener = null;
        this.lastAnchor = null;
        this.currentContext = null;
        var menuElement = Webvisum.Common.idd("contentAreaContextMenu")
        if (menuElement) {
            this.webvisum.log("installing context actions handler");
            this.listener = Webvisum.Common.objMethod(this, 'constructMenu')
            menuElement.addEventListener("popupshowing",  this.listener, false);
            return this;
        }
        return null;
    },
    destroy : function(event) {
        var menuElement = Webvisum.Common.idd("contentAreaContextMenu")
        if (menuElement && this.listener) {
            this.webvisum.log("removing context actions handler");
            menuElement.removeEventListener("popupshowing",  this.listener, false);
            this.listener = null;
        }
        this.currentContext = null;
        this.webvisum = null;
    },
    setLastAnchor:function (anAnchorElement) {
        this.webvisum.log('Setting last anchor '+anAnchorElement.href + " on document "+anAnchorElement.ownerDocument.location);
        var context = this.constructContextFromElement(anAnchorElement);
        var obj = Webvisum.Object.FactorByTypeOrName('link');
        obj.fromDOM(anAnchorElement, context);
        obj.binding = context.wdocument;
        this.lastAnchor = obj;
        this.webvisum.log("Storing last anchor:"+this.lastAnchor);
    },
    hasLastAnchor:function () {  // a context?
        return this.lastAnchor!=null;
    },
    setNewContext:function (aSource) {
        this.currentContext = new this.Context();
        this.currentContext.source = aSource;
        this.currentContext.language = this.webvisum.language;
        return this.currentContext;
    },
    isActive:function () {
        // operations are only available in Webvisum is active and user logged in
        return this.webvisum && this.webvisum.global.isActive() && this.webvisum.global.state.isLogged(); 
    },
    isBlocked:function (aTargetNode) {
        const wdoc = this.webvisum.getWDocument(aTargetNode);
        return wdoc ? wdoc.isBlocked : true;
    },
    /* constuct contenxt (and menu) - called from context menu*/
    constructMenu : function(event) {
        const hideAll = function(){ 
            gContextMenu.showItem("context_webvisum_mi_page", false);
            gContextMenu.showItem("context_webvisum_mi_link", false);
            gContextMenu.showItem("context_webvisum_mi_image", false);
            gContextMenu.showItem("context_webvisum_mi_ocr", false);
            gContextMenu.showItem("context_webvisum_mi_captcha", false);
            gContextMenu.showItem("context_webvisum_mi_form", false);
        };
        
        if(!this.isActive()){
            hideAll();
            return;
        }
        var c = this.setNewContext('contextMenu');
        if (gContextMenu) {
            const target = gContextMenu.target;
            if(this.isBlocked(target)){
                hideAll();
                return;
            }
            const isLink = c.exists.link = gContextMenu.onLink;
            const isImage = c.exists.image = gContextMenu.onImage;
            const isImageLink = isLink && isImage;
            const isForm = c.exists.form = gContextMenu.onKeywordField
                    + (target.nodeName.toLowerCase() == 'select');    
            const hasPage = c.exists.page = (c.pageURL != 'about:blank');
            const hasCaptchas = c.exists.captcha = true; //Webvisum.Monkey.captchas.length>0;

            c.doc = target.ownerDocument;
            c.pageURL = c.doc.location.toString();
            gContextMenu.showItem("context_webvisum_mi_page", hasPage);
            gContextMenu.showItem("context_webvisum_mi_link", isLink);
            gContextMenu.showItem("context_webvisum_mi_lastlink", this.hasLastAnchor());
            gContextMenu.showItem("context_webvisum_mi_image", isImage);
            gContextMenu.showItem("context_webvisum_mi_ocr", isImage);
            gContextMenu.showItem("context_webvisum_mi_captcha", hasCaptchas);
            gContextMenu.showItem("context_webvisum_mi_form", isForm);

            c.target = target;
            c.wdocument = this.webvisum.getWDocument(c.target);
            if (isLink) {
                c.linkURL = gContextMenu.linkURL;
                c.linkInfo = gContextMenu.target.innerHTML;
            }
            c.imageURL = null;
            if (isImage) {
                c.imageURL = gContextMenu.imageURL;
                c.bgImageURL = gContextMenu.bgImageURL;
                //this.imageCache(c.imageURL);
            }
        }
        this.webvisum.log(this.currentContext.toString());
    },
    //for debuging only
    imageCache:function (aUrl) {        
        this.webvisum.log("getting cache info for  '"+aUrl+"'");
        var entry = Webvisum.Monkey.fetchCacheEntry(aUrl);
        if (entry) {
            this.webvisum.log("imgInfo about '"+aUrl+"'  lastFetched: "+ new Date(1000*entry.lastFetched * 1000) );   
            this.webvisum.log("imgInfo about '"+aUrl+"' expiration: "+ new Date(1000*entry.expirationTime));   
            this.webvisum.log("dataSize: "+ entry.dataSize);   
            this.webvisum.log("device: "+ entry.deviceID);
        }    
    },
    // construct context withou ContextMenu (key shortcuts)
    constructContextFromFocus : function() {
        var c = this.setNewContext('shortcut');
        c.target = document.commandDispatcher.focusedElement;
        this._prepareContext(c);
        return c;
    },
    // called from eg. anchor onclick
    constructContextFromElement : function(anElement) {
        var c = this.setNewContext('element');
        c.target = anElement;
        this._prepareContext(c);
        return c;
    },
    _prepareContext : function(aNewContext) {
        var c = aNewContext;
        if (!c.target)
            c.target = gBrowser.selectedBrowser.contentDocument;
            
        c.doc = c.target.ownerDocument || c.target;
        if (c.doc) {
            c.pageURL = c.doc.location.toString();
            c.wdocument = this.webvisum.getWDocument(c.target);        
        }
        c.exists.link = (Webvisum.Monkey.findOuterAnchor(c.target) != null);
        c.exists.image = (c.target.nodeName.toLowerCase() == 'img');
        c.exists.captcha = true; //Webvisum.Monkey.captchas.length>0;
        c.exists.form = (c.target.nodeName.toLowerCase() == 'input') || (c.target.nodeName.toLowerCase() == 'select');
        c.exists.page = (c.pageURL != 'about:blank');
        
        if (c.exists.image) {
            c.imageURL = c.target.src;
        }
        this.webvisum.log(this.currentContext.toString());
    },
    //generic action handler
    handle : function(anAction) {
        if(!this.isActive()) return;
        if (!gContextMenu) {
            this.constructContextFromFocus();
            if (this.isBlocked(this.currentContext.target)) return;
        }
        if (this.currentContext) {
            if (anAction) 
                this['handle' + anAction].call(this, this.currentContext);
            else 
                this.handleAutomatic(this.currentContext);
        }
    },
    handleAutomatic:function (aContext) {
        this.webvisum.log('Automatic Lablel operation:' + aContext);
        if (aContext.exists.link && aContext.target) 
            this.handleLinkLabel( aContext ); 
        else if (aContext.exists.form && aContext.target)
            this.handleFormLabel( aContext ); 
        else
            this.webvisum.sound.play('error');
    },
    handleLastAnchor:function () {
        if(!this.isActive() || ! this.hasLastAnchor()) return;
        this.webvisum.log('Label last anchor operation:' + this.lastAnchor);
        var obj = this.lastAnchor;
        if ( Webvisum.Monkey.promptEditObject(obj, "you clicked last") ) {
            try{
                obj.apply(); //pass no context
                this.webvisum.cache_object(obj);
                this.webvisum.send_object(obj);
            } 
            catch(e) {
                this.webvisum.logErr("Cannot completely execute labling of "+obj, e);
            }
        }
        
    },
    handleLinkLabel : function(aContext) {
        this.webvisum.log('Label Link operation:' + aContext);
        aContext.target = Webvisum.Monkey.findOuterAnchor(aContext.target);
        if (!(aContext.exists.link && aContext.target))
            this.webvisum.sound.play('error');
        else
            this.labelSomething('link', aContext.target, aContext);
    },
    handleImageLabel : function(aContext) {        
        this.webvisum.log('Label Image operation:' + aContext);
        if (!(aContext.exists.image && aContext.target))
            this.webvisum.sound.play('error');
        else
            this.labelSomething('image', aContext.target, aContext);
    },
    handlePageLabel : function(aContext) {
        this.webvisum.log('Label Page operation:' + aContext);
        if (!(aContext.exists.page && aContext.target))
            this.webvisum.sound.play('error');
        else
            this.labelSomething('page', aContext.doc, aContext);
    },
    handleFormLabel : function(aContext) {
        this.webvisum.log('Label Form operation:' + aContext);
        if (!(aContext.exists.form && aContext.target))
            this.webvisum.sound.play('error');
        else
            this.labelSomething('form', aContext.target, aContext);
    },
    handleOcr : function(aContext) {
        if (!(aContext.exists.image && aContext.target && aContext.imageURL))
            this.webvisum.sound.play('error');
        else {
            var obj = Webvisum.Object.FactorByTypeOrName('image');
            obj.fromDOM(aContext.target, aContext);
            obj.forceTarget( aContext.target ); // a small 'helper'
            const mimeType = Webvisum.Monkey.guessMimeType(obj.id);
            this.webvisum.protocol.image_submit_request( obj, mimeType);
        }
    },
    handleCaptcha : function(aContext) {
        if (!aContext.exists.captcha)
            this.webvisum.sound.play('error');
        else {
            // find captcha image candidates
            var candidates = Webvisum.Monkey.findCaptchas(this.webvisum.global.captchaRules, aContext.doc);
            if (candidates.length==0){
	            var msg = "dialog.webvisum.captcha.notdetected";
	            this.webvisum.notifyInfo( this.webvisum.m(msg) );
              return
            }
            this.webvisum.log("Trying captcha:" +candidates[0]);
            aContext.target = candidates[0].target;
            
            var obj = Webvisum.Object.FactorByTypeOrName('image');
            obj.fromDOM(aContext.target, aContext);
            obj.forceTarget( aContext.target ); // a small 'helper'
            const mimeType = Webvisum.Monkey.guessMimeType(obj.id);
            this.webvisum.protocol.captcha_submit_request( obj, mimeType);
        }
    },
    labelSomething : function(aTypeName, anTargetElement, aContext) {
        var obj = Webvisum.Object.FactorByTypeOrName(aTypeName);
        if (obj) {
            if (obj.fromDOM(anTargetElement, aContext)!=null){
                this.webvisum.log("Promping object lable for:"+obj);
                if ( Webvisum.Monkey.promptEditObject(obj) ) {
                    try{
                        obj.apply(aContext.wdocument);
                        this.webvisum.cache_object(obj);
                        this.webvisum.send_object(obj);
                        return
                    } 
                    catch(e) {
                        this.webvisum.logErr("Cannot completely execute labling of "+obj, e);
                        this.webvisum.notifyWarning( this.webvisum.m('dialog.webvisum.label.failed') );
                        return
                    }
                }
                else return; //not labeled
            }
        }
        Webvisum.Common.logErr("Cannot factor webvisum object of the type:" + aTypeName+ '  obj:'+obj);
        this.webvisum.notifyWarning( this.webvisum.m('dialog.webvisum.label.unsupportedobject') );
    },
    getPageUrl : function() {
        alert('deprecated');
    },
    toString : function() {
        return "ContextManger, current:"+this.currentContext;
    }
};


Webvisum.ContextManager.Context = function() {
    this.doc = null;
    this.wdocument = null;
    
    this.source = null; // context menu, shortcut
    
    this.pageURL = null;
    this.linkURL = null;
    this.linkInfo = null;
    this.target = null;
    this.imageURL = null;
    this.exists = {
        link: false,
        image: false,
        captcha: false,
        form: false,
        page: false
    }
}

Webvisum.ContextManager.Context.prototype = {
      toString : function() {
        return "Context: source="+this.source+", linkUrl=" + this.linkURL + ", imageURL="
                + this.imageURL + ", linkInfo=" + this.linkInfo + ", pageURL="
                + this.pageURL + ", target="
                + (this.target ? this.target.nodeName : '')
                + Webvisum.Common.dumpObj(this.exists);
    }  
}