// REQUIRES: common.js, server.js, webvisum.js, state.js, objects.js



Webvisum.Protocol = function(aLocalWebvisum){
    this.webvisum = aLocalWebvisum;
    
    //aux methods
    this.notLogged = function(optMsg) {
        if (!this.state.isLogged()) {
            this.webvisum.log(optMsg + ' skipping. Not logged in')
            return true
        }
        return false
    };
    this.common_response = function (aResponse, aContext) {
        if (aResponse.hasServerMsg()) {
            this.webvisum.fetch_messages();
        }
    };
}

Webvisum.Protocol.prototype = {
    API_VERSION : '1.0',
    MAX_POLLING_ATTEMPTS : 10,
    MAX_POLLING_TIMEOUT : 120, //seconds
    
    get server(){
        return this.webvisum.server  
    },    
    get state(){
        return this.webvisum.global.state  
    },    
    /* 
     * login 
     */
    auth_login_request : function(optDelayMs, optFailureHandler) {
        this.webvisum.log('sending auth.login');
        var rq = this.server.newRequest('auth.login');
        rq.addParams({
            api_version : this.API_VERSION,
            app_version : this.webvisum.global.APP_VERSION,
            user_name : rq.username,
            password : rq.pwd
        });
        rq.setResponseHandler(this, 'auth_login_response', 'auth_login_failure' );
        const context = {
            failureHandler : optFailureHandler
        };
        this.server.sendRequest(rq, context, {delay: 1*optDelayMs});
    },
    auth_login_response : function(aResponse, aContext) {
        this.common_response(aResponse); 
        const session_key = aResponse.tagText('session_key');
        const last_captcha_rules = 1*aResponse.tagText('captcha_rules_last_update');
        this.state.session_key = session_key;
        const last_cached_rules = this.webvisum.global.captchaRules.last_update; 
        this.webvisum.log('Logged in! Session:' + session_key 
            + ", server captcha:"+last_captcha_rules+", cached captcha:"
            + last_cached_rules);
      
        // handle captcha rules here
        if (last_captcha_rules > last_cached_rules){
            this.webvisum.log('Captcha rules needs an update');
            this.captcha_getrules_request( last_cached_rules, null, 2);
        }
    },
    auth_login_failure : function(aResponse, aContext) {
        this.common_response(aResponse); 
        var errMgs  = aResponse.getErrorText();
        this.webvisum.logWrn('Logging failure. ' + errMgs);
      
        if (aResponse.getErrorCode()==12){
            //account registration not completed yet
            this.webvisum.notifyWarning( this.webvisum.m("dialog.webvisum.login.registration_pending" ) );
        }
        else if (aContext.failureHandler) {
            aContext.failureHandler(true);            
        }
    },
    /* 
     * logout 
     */
    auth_logout_request : function() {
        this.webvisum.log('sending auth.logout');
        if (!this.state.session_key) return;
        var rq = this.server.newRequest('auth.logout', {
            session_key : this.state.session_key
        });
        rq.setResponseHandler(this, 'auth_logout_response');
        //apply local logout immediately
        this.state.session_key = null;
        this.server.sendRequest(rq);
    },
    auth_logout_response : function(aResponse) {
        //prevent the "logged out" dialog from appearing twice when logging out
        if (this.state.session_key !== null){
          this.state.session_key = null;
          this.webvisum.log('Logged out');
        }
    },
    /* 
     * domain.info
     * anObjectsHandler  = fun(aObjsAry, aDomain, aTimestamp, aPageUrl, aLanguage) 
     */
    domains_getinfo_request : function( aPageUrl, aDomain, anObjectsHandler, 
            aLastUpdate, aDelayTimeMs) {
        if (this.notLogged('domains.getinfo')) return;                
        this.webvisum.log('sending domains.getinfo');
        var rq = this.server.newRequest('domains.getinfo', {
            session_key : this.state.session_key,
            language : this.webvisum.language,
            compressed : 0,
            last_update: aLastUpdate,
            page_url : aPageUrl
        });
        rq.setResponseHandler(this, 'domains_getinfo_response');
        const key = "<domains.getinfo.global."+aDomain+">";
        const context = {
            pageUrl : aPageUrl,
            handler : anObjectsHandler
        };
        rq.setKey(key);
        this.server.queueRequest(rq, context, {
            delay: aDelayTimeMs, 
            frequencyGuard: (aDomain ? 60*1000 : -1) });
    },
    domains_getinfo_response : function(aResponse, aContext) {
        this.common_response(aResponse, aContext);
        this.webvisum.log('Received objects info: ' + aResponse.root + "  ctx:" + aContext);
        const last_update = aResponse.tagText('last_update');
        const elems = Webvisum.Common.tags('object', aResponse.root);
        const domain = Webvisum.Common.tag('domain', aResponse.root).textContent;
        const lang = Webvisum.Common.tag('language', aResponse.root).textContent;
        const pageUrl = aContext.pageUrl;
        var objs = [];
        for (var i = 0; i < elems.length; i++) {
            var obj = Webvisum.Object.FactorFromXML(elems[i], pageUrl, lang);
            if (obj) {
                objs.push(obj);
            }
        }
        this.webvisum.log("Response domain:'" + domain + "' pageUrl:" + pageUrl + " lang:"
                + lang + " update:" + last_update +"="+Webvisum.Common.toDate(last_update)+ " objs:" + objs.length);
        aContext.handler(objs, domain, last_update, pageUrl, lang);
    },
    /* 
     * objects.label 
     */
    objects_label_request : function( aWebvisumObject) {
        if (this.notLogged('objects.label')) return;
        this.webvisum.log('sending objects.label:' + aWebvisumObject);
        var rq = this.server.newRequest('objects.label', {
            session_key : this.state.session_key,
            language : aWebvisumObject.language,
            page_url : aWebvisumObject.page_url,
            object_type : aWebvisumObject.type,
            object_id : aWebvisumObject.id,
            label : aWebvisumObject.label
        });
        rq.setResponseHandler(this, 'objects_label_response');
        const key = "<" + aWebvisumObject.name + "_" + aWebvisumObject.page_url + "_"
                + aWebvisumObject.id + "_"+aWebvisumObject.language+">";
        rq.setKey(key);
        this.server.queueRequest(rq, aWebvisumObject, {delay: 200});
    },
    objects_label_response : function(aResponse, aWebvisumObject) {
        this.common_response(aResponse);
        aWebvisumObject.isDirtyServer = false;
        if (this.webvisum.isVerbose()) {
            msg = "dialog.webvisum.label.submitted";
            this.webvisum.notifyInfo( this.webvisum.m(msg) );
        }
        //NO OP
        this.webvisum.log('Received label object confirm');
    },

    /* 
     * image.* 
     */
    image_submit_request : function(anObject, aMimeType) {
        if (this.notLogged('image.submit')) return;
        this.webvisum.log('sending image.submit:' + anObject + " mimetype:" + aMimeType);
        //prepare image data
        const canvas = Webvisum.Monkey.image2canvas(anObject.target);
        const imageUri = Webvisum.Monkey.canvas2uri(canvas);
        //prepare request
        var rq = this.server.newRequest('image.submit', {
            session_key : this.state.session_key,
            //        language: webvisum.language,
            page_url : anObject.page_url,
            object_id : anObject.id
        }, Webvisum.UploadRequest.initFrom(this.server.genericRequest));
        rq.setResourceUri(imageUri, aMimeType, 'image');
        rq.setResponseHandler(this, 'image_response');
        const context = {
            object : anObject,
            attempt: 0,
            started_on: (1*new Date())
        };
        this.server.sendRequest(rq, context);
        if (this.webvisum.isVerbose()) {
            msg = "dialog.webvisum.ocr.submit";
            this.webvisum.notifyInfo( this.webvisum.m(msg) );
        }
    },
    image_result_request : function( aHash, aDelaySec, aContext) {
        if (this.notLogged('image.result')) return;
        aContext.attempt+=1;
        var timediff = Math.round((1*new Date()-aContext.started_on)/1000);
        if (aContext.attempt > this.MAX_POLLING_ATTEMPTS || timediff > this.MAX_POLLING_TIMEOUT) {
            this.webvisum.logErr("Maximum polling attempts "+this.MAX_POLLING_ATTEMPTS+ " or timeout "+this.MAX_POLLING_TIMEOUT+"sec exceeded.");
            this.webvisum.notifyWarning( this.webvisum.m( "dialog.webvisum.ocr.timeout" ) );
            return
        }
        this.webvisum.log('sending image.result  hash:' + aHash + ', eta: ' + aDelaySec);
        const rq = this.server.newRequest('image.result', {
            session_key : this.state.session_key,
            hash : aHash
        });
        rq.setResponseHandler(this, 'image_response');
        var object = aContext.object;
        const key = "<ocr_" + object.page_url + "_" + object.id + ">";
        rq.setKey(key);
        this.server.queueRequest(rq, aContext, 
            {delay: aDelaySec * 1000});
    },
    image_response : function(aResponse, aContext) {
        this.common_response(aResponse, aContext);
        var timediff = Math.round((1*new Date()-aContext.started_on)/1000);
        this.webvisum.log('Received image solving response, elapsed time:'+timediff+" sec");
        const hash = aResponse.tagText('hash');
        const eta = aResponse.tagText('eta');
        const text = aResponse.tagText('text');
        this.webvisum.log('Image solver eta:' + eta + ' sec, hash:' + hash + ", text:" + text);
        if (eta > 0) {
            //wait and retry
            this.image_result_request( hash, eta, aContext);
        } else if (eta == 0) {
            this.webvisum.notifyInfo( this.webvisum.m("dialog.webvisum.ocr.solved" , [text]) ); 
            //update image            
            aContext.object.editLabel( text );
            try {            
                aContext.object.apply();
            } 
            catch(e) {
                this.webvisum.logErr("Cannot completely execute labling of image "+obj, e);
            }

        } else {
            //not in queue
            this.webvisum.notifyWarning( this.webvisum.m( "dialog.webvisum.ocr.notresolved", [aContext.object.id] ) );
        }
    },
    /* 
     * captcha.* 
     */
    captcha_submit_request : function( anObject, aMimeType) {
        if (this.notLogged('captcha.submit')) return;
        this.webvisum.log('sending captcha.submit:' + anObject + " mimetype:" + aMimeType);  
        //prepare image data
        const canvas = Webvisum.Monkey.image2canvas(anObject.target);
        const imageUri = Webvisum.Monkey.canvas2uri(canvas);
        //prepare request

        var rq = this.server.newRequest('captcha.submit', {
            session_key : this.state.session_key,
            page_url : anObject.page_url
        }, Webvisum.UploadRequest.initFrom(this.server.genericRequest));
        rq.setResourceUri(imageUri, aMimeType, 'captcha');
        rq.setResponseHandler(this, 'captcha_response');
        const context = {
            object : anObject,
            attempt: 0,
            started_on: (1*new Date())
        };
        this.server.sendRequest(rq, context);
        if (this.webvisum.isVerbose()) {
            var msg = "dialog.webvisum.captcha.submit";
            this.webvisum.notifyInfo( this.webvisum.m(msg) );
        }

    },
    captcha_result_request : function(aHash, aDelaySec, aContext) {
        if (this.notLogged('captcha.result')) return;
        aContext.attempt+=1;
        var timediff = Math.round((1*new Date()-aContext.started_on)/1000);
        if (aContext.attempt > this.MAX_POLLING_ATTEMPTS || timediff > this.MAX_POLLING_TIMEOUT) {
            this.webvisum.logErr("Maximum polling attempts "+this.MAX_POLLING_ATTEMPTS+ " or timeout "+this.MAX_POLLING_TIMEOUT+"sec exceeded.");
            this.webvisum.notifyWarning( this.webvisum.m( "dialog.webvisum.captcha.timeout" ) );
            return
        }
        
        this.webvisum.log('sending captcha.result hash:' + aHash + ', eta: ' + aDelaySec);
        const rq = this.server.newRequest('captcha.result', {
            session_key : this.state.session_key,
            hash : aHash
        });
        rq.setResponseHandler(this, 'captcha_response');
        var object = aContext.object;
        const key = "<captcha_" + object.page_url + "_" + object.id + ">";
        rq.setKey(key);
        this.server.queueRequest(rq, aContext, 
            {delay: aDelaySec * 1000});

    },
    captcha_response : function(aResponse, aContext) {
        this.common_response(aResponse, aContext);
        var timediff = Math.round((1*new Date()-aContext.started_on)/1000);
        this.webvisum.log('Received captcha solving response, elapsed time:'+timediff+" sec");
        const hash = aResponse.tagText('hash');
        const eta = aResponse.tagText('eta');
        const text = aResponse.tagText('text');
        this.webvisum.log('Captcha solver eta:' + eta + ' sec, hash:' + hash + ", text:"
                + text);
        if (eta > 0) {
            //wait and retry
            this.captcha_result_request( hash, eta, aContext);
        } else if ((eta < 0) || !(text && text.length > 0)) {
            //not in queue OR no solution
            this.webvisum.notifyWarning( this.webvisum.m("dialog.webvisum.captcha.notsolved") );
        } else {
            Webvisum.Monkey.copyToClipboard(text);
            const domain = Webvisum.Monkey
                    .extractDomainFromUrl(aContext.object.page_url);
            this.webvisum.notifyInfo( this.webvisum.m("dialog.webvisum.captcha.solved" , [domain, text]) ); 
//            alert( this.webvisum.m("dialog.webvisum.captcha.solved" , [domain, text]) ); 
        }
    },
    /*
     * captcha.getrules
     * aRulesHandler = callback handler(rules, last_update)
     */
     captcha_getrules_request : function( aLastUpdate, optRulesHandler, optDelaySec) {
        if (this.notLogged('captcha.getrules')) return;
        this.webvisum.log('sending captcha.getrules, last update ' + aLastUpdate);
        const rq = this.server.newRequest('captcha.getrules', {
            session_key : this.state.session_key,
            last_update : aLastUpdate
        });
        rq.setResponseHandler(this, 'captcha_getrules_response');
        const key = "<captcharules>";
        var context = null;
        if (optRulesHandler){
            context = { handler : optRulesHandler };
        }
        rq.setKey(key);
        this.server.queueRequest(rq, context, 
            {delay: optDelaySec * 1000});

    },
    captcha_getrules_response : function(aResponse, aContext) {
        this.common_response(aResponse, aContext);
        this.webvisum.log('Received captcha rules response');
        var captchaRules = Webvisum.beget(Webvisum.CaptchaRules).fromXml(aResponse.root.ownerDocument);
        if (captchaRules){
            this.webvisum.log('Parsed '+captchaRules.rules.length+' rules'); 
            this.webvisum.global.updateCaptchaRules(captchaRules);    
        }
        if (aContext) aContext.handler(captchaRules);
    },
    
    /* 
     * config.load 
     */
    config_load_request : function() {
        if (this.notLogged('config.load')) return;
        this.webvisum.log('sending config.load');
        var rq = this.server.newRequest('config.load', {
            session_key : this.state.session_key
        });
        rq.setResponseHandler(this, 'config_load_response');
        this.server.sendRequest(rq);
    },
    config_load_response : function(aResponse) {
        this.common_response(aResponse);
        const config = aResponse.tagText('config')
        //TODO: config
        this.webvisum.log('Received config_load: '+config);
    },
    /* 
     * config.save 
     */
    config_save_request : function() {
        if (this.notLogged('config.save')) return;
        this.webvisum.log('sending config.save');
        var rq = this.server.newRequest('config.save', {
            session_key : this.state.session_key,
            config : 'a=123'
        });
        rq.setResponseHandler(this, 'config_save_response');
        this.server.sendRequest(rq);
    },
    config_save_response : function(aResponse) {
        this.common_response(aResponse);
        // NO OP
        this.webvisum.log('Received config_save');
    },
    
    /* 
     * message.read - Fetches server messages
     */
    messages_read : function(optDelayMs) {
        if (this.notLogged('message.read')) return;
        this.webvisum.log('sending messages.read');
        optDelayMs = (optDelayMs == null) ? 1000 : optDelayMs;
        var rq = this.server.newRequest('messages.read');
        rq.addParams({
            api_version : this.API_VERSION,
            app_version : this.webvisum.global.APP_VERSION,
            session_key : this.state.session_key
        });
        rq.setResponseHandler(this, 'messages_read_response');
        this.server.queueRequest(rq, null, 
            {submissive:true, delay: optDelayMs});
    },
    messages_read_response : function(aResponse) {
        this.webvisum.notifyInfo( this.webvisum.m("menu.webvisum.msg.servermsg")+": \n"+aResponse.tagText('message') );
    }
}